/* Copyright (c) 2015, Brandon Jones, Colin MacKenzie IV.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE. */

if(!GLMAT_EPSILON) {
    var GLMAT_EPSILON = 0.000001;
}

if(!GLMAT_ARRAY_TYPE) {
    var GLMAT_ARRAY_TYPE = (typeof Float32Array !== 'undefined') ? Float32Array : Array;
}

if(!GLMAT_RANDOM) {
    var GLMAT_RANDOM = Math.random;
}

/**
 * @class Common utilities
 * @name glMatrix
 */
var glMatrix = {};

/**
 * Sets the type of array used when creating new vectors and matrices
 *
 * @param {Type} type Array type, such as Float32Array or Array
 */
glMatrix.setMatrixArrayType = function(type) {
    GLMAT_ARRAY_TYPE = type;
}

if(typeof(exports) !== 'undefined') {
    exports.glMatrix = glMatrix;
}

var degree = Math.PI / 180;

/**
* Convert Degree To Radian
*
* @param {Number} Angle in Degrees
*/
glMatrix.toRadian = function(a){
     return a * degree;
}
/* Copyright (c) 2015, Brandon Jones, Colin MacKenzie IV.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE. */

/**
 * @class 4x4 Matrix
 * @name mat4
 */
var mat4 = {};

/**
 * Creates a new identity mat4
 *
 * @returns {mat4} a new 4x4 matrix
 */
mat4.create = function() {
    var out = new GLMAT_ARRAY_TYPE(16);
    out[0] = 1;
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
    out[4] = 0;
    out[5] = 1;
    out[6] = 0;
    out[7] = 0;
    out[8] = 0;
    out[9] = 0;
    out[10] = 1;
    out[11] = 0;
    out[12] = 0;
    out[13] = 0;
    out[14] = 0;
    out[15] = 1;
    return out;
};

/**
 * Creates a new mat4 initialized with values from an existing matrix
 *
 * @param {mat4} a matrix to clone
 * @returns {mat4} a new 4x4 matrix
 */
mat4.clone = function(a) {
    var out = new GLMAT_ARRAY_TYPE(16);
    out[0] = a[0];
    out[1] = a[1];
    out[2] = a[2];
    out[3] = a[3];
    out[4] = a[4];
    out[5] = a[5];
    out[6] = a[6];
    out[7] = a[7];
    out[8] = a[8];
    out[9] = a[9];
    out[10] = a[10];
    out[11] = a[11];
    out[12] = a[12];
    out[13] = a[13];
    out[14] = a[14];
    out[15] = a[15];
    return out;
};

/**
 * Copy the values from one mat4 to another
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the source matrix
 * @returns {mat4} out
 */
mat4.copy = function(out, a) {
    out[0] = a[0];
    out[1] = a[1];
    out[2] = a[2];
    out[3] = a[3];
    out[4] = a[4];
    out[5] = a[5];
    out[6] = a[6];
    out[7] = a[7];
    out[8] = a[8];
    out[9] = a[9];
    out[10] = a[10];
    out[11] = a[11];
    out[12] = a[12];
    out[13] = a[13];
    out[14] = a[14];
    out[15] = a[15];
    return out;
};

/**
 * Set a mat4 to the identity matrix
 *
 * @param {mat4} out the receiving matrix
 * @returns {mat4} out
 */
mat4.identity = function(out) {
    out[0] = 1;
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
    out[4] = 0;
    out[5] = 1;
    out[6] = 0;
    out[7] = 0;
    out[8] = 0;
    out[9] = 0;
    out[10] = 1;
    out[11] = 0;
    out[12] = 0;
    out[13] = 0;
    out[14] = 0;
    out[15] = 1;
    return out;
};

/**
 * Transpose the values of a mat4
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the source matrix
 * @returns {mat4} out
 */
mat4.transpose = function(out, a) {
    // If we are transposing ourselves we can skip a few steps but have to cache some values
    if (out === a) {
        var a01 = a[1], a02 = a[2], a03 = a[3],
            a12 = a[6], a13 = a[7],
            a23 = a[11];

        out[1] = a[4];
        out[2] = a[8];
        out[3] = a[12];
        out[4] = a01;
        out[6] = a[9];
        out[7] = a[13];
        out[8] = a02;
        out[9] = a12;
        out[11] = a[14];
        out[12] = a03;
        out[13] = a13;
        out[14] = a23;
    } else {
        out[0] = a[0];
        out[1] = a[4];
        out[2] = a[8];
        out[3] = a[12];
        out[4] = a[1];
        out[5] = a[5];
        out[6] = a[9];
        out[7] = a[13];
        out[8] = a[2];
        out[9] = a[6];
        out[10] = a[10];
        out[11] = a[14];
        out[12] = a[3];
        out[13] = a[7];
        out[14] = a[11];
        out[15] = a[15];
    }
    
    return out;
};

/**
 * Inverts a mat4
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the source matrix
 * @returns {mat4} out
 */
mat4.invert = function(out, a) {
    var a00 = a[0], a01 = a[1], a02 = a[2], a03 = a[3],
        a10 = a[4], a11 = a[5], a12 = a[6], a13 = a[7],
        a20 = a[8], a21 = a[9], a22 = a[10], a23 = a[11],
        a30 = a[12], a31 = a[13], a32 = a[14], a33 = a[15],

        b00 = a00 * a11 - a01 * a10,
        b01 = a00 * a12 - a02 * a10,
        b02 = a00 * a13 - a03 * a10,
        b03 = a01 * a12 - a02 * a11,
        b04 = a01 * a13 - a03 * a11,
        b05 = a02 * a13 - a03 * a12,
        b06 = a20 * a31 - a21 * a30,
        b07 = a20 * a32 - a22 * a30,
        b08 = a20 * a33 - a23 * a30,
        b09 = a21 * a32 - a22 * a31,
        b10 = a21 * a33 - a23 * a31,
        b11 = a22 * a33 - a23 * a32,

        // Calculate the determinant
        det = b00 * b11 - b01 * b10 + b02 * b09 + b03 * b08 - b04 * b07 + b05 * b06;

    if (!det) { 
        return null; 
    }
    det = 1.0 / det;

    out[0] = (a11 * b11 - a12 * b10 + a13 * b09) * det;
    out[1] = (a02 * b10 - a01 * b11 - a03 * b09) * det;
    out[2] = (a31 * b05 - a32 * b04 + a33 * b03) * det;
    out[3] = (a22 * b04 - a21 * b05 - a23 * b03) * det;
    out[4] = (a12 * b08 - a10 * b11 - a13 * b07) * det;
    out[5] = (a00 * b11 - a02 * b08 + a03 * b07) * det;
    out[6] = (a32 * b02 - a30 * b05 - a33 * b01) * det;
    out[7] = (a20 * b05 - a22 * b02 + a23 * b01) * det;
    out[8] = (a10 * b10 - a11 * b08 + a13 * b06) * det;
    out[9] = (a01 * b08 - a00 * b10 - a03 * b06) * det;
    out[10] = (a30 * b04 - a31 * b02 + a33 * b00) * det;
    out[11] = (a21 * b02 - a20 * b04 - a23 * b00) * det;
    out[12] = (a11 * b07 - a10 * b09 - a12 * b06) * det;
    out[13] = (a00 * b09 - a01 * b07 + a02 * b06) * det;
    out[14] = (a31 * b01 - a30 * b03 - a32 * b00) * det;
    out[15] = (a20 * b03 - a21 * b01 + a22 * b00) * det;

    return out;
};

/**
 * Calculates the adjugate of a mat4
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the source matrix
 * @returns {mat4} out
 */
mat4.adjoint = function(out, a) {
    var a00 = a[0], a01 = a[1], a02 = a[2], a03 = a[3],
        a10 = a[4], a11 = a[5], a12 = a[6], a13 = a[7],
        a20 = a[8], a21 = a[9], a22 = a[10], a23 = a[11],
        a30 = a[12], a31 = a[13], a32 = a[14], a33 = a[15];

    out[0]  =  (a11 * (a22 * a33 - a23 * a32) - a21 * (a12 * a33 - a13 * a32) + a31 * (a12 * a23 - a13 * a22));
    out[1]  = -(a01 * (a22 * a33 - a23 * a32) - a21 * (a02 * a33 - a03 * a32) + a31 * (a02 * a23 - a03 * a22));
    out[2]  =  (a01 * (a12 * a33 - a13 * a32) - a11 * (a02 * a33 - a03 * a32) + a31 * (a02 * a13 - a03 * a12));
    out[3]  = -(a01 * (a12 * a23 - a13 * a22) - a11 * (a02 * a23 - a03 * a22) + a21 * (a02 * a13 - a03 * a12));
    out[4]  = -(a10 * (a22 * a33 - a23 * a32) - a20 * (a12 * a33 - a13 * a32) + a30 * (a12 * a23 - a13 * a22));
    out[5]  =  (a00 * (a22 * a33 - a23 * a32) - a20 * (a02 * a33 - a03 * a32) + a30 * (a02 * a23 - a03 * a22));
    out[6]  = -(a00 * (a12 * a33 - a13 * a32) - a10 * (a02 * a33 - a03 * a32) + a30 * (a02 * a13 - a03 * a12));
    out[7]  =  (a00 * (a12 * a23 - a13 * a22) - a10 * (a02 * a23 - a03 * a22) + a20 * (a02 * a13 - a03 * a12));
    out[8]  =  (a10 * (a21 * a33 - a23 * a31) - a20 * (a11 * a33 - a13 * a31) + a30 * (a11 * a23 - a13 * a21));
    out[9]  = -(a00 * (a21 * a33 - a23 * a31) - a20 * (a01 * a33 - a03 * a31) + a30 * (a01 * a23 - a03 * a21));
    out[10] =  (a00 * (a11 * a33 - a13 * a31) - a10 * (a01 * a33 - a03 * a31) + a30 * (a01 * a13 - a03 * a11));
    out[11] = -(a00 * (a11 * a23 - a13 * a21) - a10 * (a01 * a23 - a03 * a21) + a20 * (a01 * a13 - a03 * a11));
    out[12] = -(a10 * (a21 * a32 - a22 * a31) - a20 * (a11 * a32 - a12 * a31) + a30 * (a11 * a22 - a12 * a21));
    out[13] =  (a00 * (a21 * a32 - a22 * a31) - a20 * (a01 * a32 - a02 * a31) + a30 * (a01 * a22 - a02 * a21));
    out[14] = -(a00 * (a11 * a32 - a12 * a31) - a10 * (a01 * a32 - a02 * a31) + a30 * (a01 * a12 - a02 * a11));
    out[15] =  (a00 * (a11 * a22 - a12 * a21) - a10 * (a01 * a22 - a02 * a21) + a20 * (a01 * a12 - a02 * a11));
    return out;
};

/**
 * Calculates the determinant of a mat4
 *
 * @param {mat4} a the source matrix
 * @returns {Number} determinant of a
 */
mat4.determinant = function (a) {
    var a00 = a[0], a01 = a[1], a02 = a[2], a03 = a[3],
        a10 = a[4], a11 = a[5], a12 = a[6], a13 = a[7],
        a20 = a[8], a21 = a[9], a22 = a[10], a23 = a[11],
        a30 = a[12], a31 = a[13], a32 = a[14], a33 = a[15],

        b00 = a00 * a11 - a01 * a10,
        b01 = a00 * a12 - a02 * a10,
        b02 = a00 * a13 - a03 * a10,
        b03 = a01 * a12 - a02 * a11,
        b04 = a01 * a13 - a03 * a11,
        b05 = a02 * a13 - a03 * a12,
        b06 = a20 * a31 - a21 * a30,
        b07 = a20 * a32 - a22 * a30,
        b08 = a20 * a33 - a23 * a30,
        b09 = a21 * a32 - a22 * a31,
        b10 = a21 * a33 - a23 * a31,
        b11 = a22 * a33 - a23 * a32;

    // Calculate the determinant
    return b00 * b11 - b01 * b10 + b02 * b09 + b03 * b08 - b04 * b07 + b05 * b06;
};

/**
 * Multiplies two mat4's
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the first operand
 * @param {mat4} b the second operand
 * @returns {mat4} out
 */
mat4.multiply = function (out, a, b) {
    var a00 = a[0], a01 = a[1], a02 = a[2], a03 = a[3],
        a10 = a[4], a11 = a[5], a12 = a[6], a13 = a[7],
        a20 = a[8], a21 = a[9], a22 = a[10], a23 = a[11],
        a30 = a[12], a31 = a[13], a32 = a[14], a33 = a[15];

    // Cache only the current line of the second matrix
    var b0  = b[0], b1 = b[1], b2 = b[2], b3 = b[3];  
    out[0] = b0*a00 + b1*a10 + b2*a20 + b3*a30;
    out[1] = b0*a01 + b1*a11 + b2*a21 + b3*a31;
    out[2] = b0*a02 + b1*a12 + b2*a22 + b3*a32;
    out[3] = b0*a03 + b1*a13 + b2*a23 + b3*a33;

    b0 = b[4]; b1 = b[5]; b2 = b[6]; b3 = b[7];
    out[4] = b0*a00 + b1*a10 + b2*a20 + b3*a30;
    out[5] = b0*a01 + b1*a11 + b2*a21 + b3*a31;
    out[6] = b0*a02 + b1*a12 + b2*a22 + b3*a32;
    out[7] = b0*a03 + b1*a13 + b2*a23 + b3*a33;

    b0 = b[8]; b1 = b[9]; b2 = b[10]; b3 = b[11];
    out[8] = b0*a00 + b1*a10 + b2*a20 + b3*a30;
    out[9] = b0*a01 + b1*a11 + b2*a21 + b3*a31;
    out[10] = b0*a02 + b1*a12 + b2*a22 + b3*a32;
    out[11] = b0*a03 + b1*a13 + b2*a23 + b3*a33;

    b0 = b[12]; b1 = b[13]; b2 = b[14]; b3 = b[15];
    out[12] = b0*a00 + b1*a10 + b2*a20 + b3*a30;
    out[13] = b0*a01 + b1*a11 + b2*a21 + b3*a31;
    out[14] = b0*a02 + b1*a12 + b2*a22 + b3*a32;
    out[15] = b0*a03 + b1*a13 + b2*a23 + b3*a33;
    return out;
};

/**
 * Alias for {@link mat4.multiply}
 * @function
 */
mat4.mul = mat4.multiply;

/**
 * Translate a mat4 by the given vector
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the matrix to translate
 * @param {vec3} v vector to translate by
 * @returns {mat4} out
 */
mat4.translate = function (out, a, v) {
    var x = v[0], y = v[1], z = v[2],
        a00, a01, a02, a03,
        a10, a11, a12, a13,
        a20, a21, a22, a23;

    if (a === out) {
        out[12] = a[0] * x + a[4] * y + a[8] * z + a[12];
        out[13] = a[1] * x + a[5] * y + a[9] * z + a[13];
        out[14] = a[2] * x + a[6] * y + a[10] * z + a[14];
        out[15] = a[3] * x + a[7] * y + a[11] * z + a[15];
    } else {
        a00 = a[0]; a01 = a[1]; a02 = a[2]; a03 = a[3];
        a10 = a[4]; a11 = a[5]; a12 = a[6]; a13 = a[7];
        a20 = a[8]; a21 = a[9]; a22 = a[10]; a23 = a[11];

        out[0] = a00; out[1] = a01; out[2] = a02; out[3] = a03;
        out[4] = a10; out[5] = a11; out[6] = a12; out[7] = a13;
        out[8] = a20; out[9] = a21; out[10] = a22; out[11] = a23;

        out[12] = a00 * x + a10 * y + a20 * z + a[12];
        out[13] = a01 * x + a11 * y + a21 * z + a[13];
        out[14] = a02 * x + a12 * y + a22 * z + a[14];
        out[15] = a03 * x + a13 * y + a23 * z + a[15];
    }

    return out;
};

/**
 * Scales the mat4 by the dimensions in the given vec3
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the matrix to scale
 * @param {vec3} v the vec3 to scale the matrix by
 * @returns {mat4} out
 **/
mat4.scale = function(out, a, v) {
    var x = v[0], y = v[1], z = v[2];

    out[0] = a[0] * x;
    out[1] = a[1] * x;
    out[2] = a[2] * x;
    out[3] = a[3] * x;
    out[4] = a[4] * y;
    out[5] = a[5] * y;
    out[6] = a[6] * y;
    out[7] = a[7] * y;
    out[8] = a[8] * z;
    out[9] = a[9] * z;
    out[10] = a[10] * z;
    out[11] = a[11] * z;
    out[12] = a[12];
    out[13] = a[13];
    out[14] = a[14];
    out[15] = a[15];
    return out;
};

/**
 * Rotates a mat4 by the given angle around the given axis
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the matrix to rotate
 * @param {Number} rad the angle to rotate the matrix by
 * @param {vec3} axis the axis to rotate around
 * @returns {mat4} out
 */
mat4.rotate = function (out, a, rad, axis) {
    var x = axis[0], y = axis[1], z = axis[2],
        len = Math.sqrt(x * x + y * y + z * z),
        s, c, t,
        a00, a01, a02, a03,
        a10, a11, a12, a13,
        a20, a21, a22, a23,
        b00, b01, b02,
        b10, b11, b12,
        b20, b21, b22;

    if (Math.abs(len) < GLMAT_EPSILON) { return null; }
    
    len = 1 / len;
    x *= len;
    y *= len;
    z *= len;

    s = Math.sin(rad);
    c = Math.cos(rad);
    t = 1 - c;

    a00 = a[0]; a01 = a[1]; a02 = a[2]; a03 = a[3];
    a10 = a[4]; a11 = a[5]; a12 = a[6]; a13 = a[7];
    a20 = a[8]; a21 = a[9]; a22 = a[10]; a23 = a[11];

    // Construct the elements of the rotation matrix
    b00 = x * x * t + c; b01 = y * x * t + z * s; b02 = z * x * t - y * s;
    b10 = x * y * t - z * s; b11 = y * y * t + c; b12 = z * y * t + x * s;
    b20 = x * z * t + y * s; b21 = y * z * t - x * s; b22 = z * z * t + c;

    // Perform rotation-specific matrix multiplication
    out[0] = a00 * b00 + a10 * b01 + a20 * b02;
    out[1] = a01 * b00 + a11 * b01 + a21 * b02;
    out[2] = a02 * b00 + a12 * b01 + a22 * b02;
    out[3] = a03 * b00 + a13 * b01 + a23 * b02;
    out[4] = a00 * b10 + a10 * b11 + a20 * b12;
    out[5] = a01 * b10 + a11 * b11 + a21 * b12;
    out[6] = a02 * b10 + a12 * b11 + a22 * b12;
    out[7] = a03 * b10 + a13 * b11 + a23 * b12;
    out[8] = a00 * b20 + a10 * b21 + a20 * b22;
    out[9] = a01 * b20 + a11 * b21 + a21 * b22;
    out[10] = a02 * b20 + a12 * b21 + a22 * b22;
    out[11] = a03 * b20 + a13 * b21 + a23 * b22;

    if (a !== out) { // If the source and destination differ, copy the unchanged last row
        out[12] = a[12];
        out[13] = a[13];
        out[14] = a[14];
        out[15] = a[15];
    }
    return out;
};

/**
 * Rotates a matrix by the given angle around the X axis
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the matrix to rotate
 * @param {Number} rad the angle to rotate the matrix by
 * @returns {mat4} out
 */
mat4.rotateX = function (out, a, rad) {
    var s = Math.sin(rad),
        c = Math.cos(rad),
        a10 = a[4],
        a11 = a[5],
        a12 = a[6],
        a13 = a[7],
        a20 = a[8],
        a21 = a[9],
        a22 = a[10],
        a23 = a[11];

    if (a !== out) { // If the source and destination differ, copy the unchanged rows
        out[0]  = a[0];
        out[1]  = a[1];
        out[2]  = a[2];
        out[3]  = a[3];
        out[12] = a[12];
        out[13] = a[13];
        out[14] = a[14];
        out[15] = a[15];
    }

    // Perform axis-specific matrix multiplication
    out[4] = a10 * c + a20 * s;
    out[5] = a11 * c + a21 * s;
    out[6] = a12 * c + a22 * s;
    out[7] = a13 * c + a23 * s;
    out[8] = a20 * c - a10 * s;
    out[9] = a21 * c - a11 * s;
    out[10] = a22 * c - a12 * s;
    out[11] = a23 * c - a13 * s;
    return out;
};

/**
 * Rotates a matrix by the given angle around the Y axis
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the matrix to rotate
 * @param {Number} rad the angle to rotate the matrix by
 * @returns {mat4} out
 */
mat4.rotateY = function (out, a, rad) {
    var s = Math.sin(rad),
        c = Math.cos(rad),
        a00 = a[0],
        a01 = a[1],
        a02 = a[2],
        a03 = a[3],
        a20 = a[8],
        a21 = a[9],
        a22 = a[10],
        a23 = a[11];

    if (a !== out) { // If the source and destination differ, copy the unchanged rows
        out[4]  = a[4];
        out[5]  = a[5];
        out[6]  = a[6];
        out[7]  = a[7];
        out[12] = a[12];
        out[13] = a[13];
        out[14] = a[14];
        out[15] = a[15];
    }

    // Perform axis-specific matrix multiplication
    out[0] = a00 * c - a20 * s;
    out[1] = a01 * c - a21 * s;
    out[2] = a02 * c - a22 * s;
    out[3] = a03 * c - a23 * s;
    out[8] = a00 * s + a20 * c;
    out[9] = a01 * s + a21 * c;
    out[10] = a02 * s + a22 * c;
    out[11] = a03 * s + a23 * c;
    return out;
};

/**
 * Rotates a matrix by the given angle around the Z axis
 *
 * @param {mat4} out the receiving matrix
 * @param {mat4} a the matrix to rotate
 * @param {Number} rad the angle to rotate the matrix by
 * @returns {mat4} out
 */
mat4.rotateZ = function (out, a, rad) {
    var s = Math.sin(rad),
        c = Math.cos(rad),
        a00 = a[0],
        a01 = a[1],
        a02 = a[2],
        a03 = a[3],
        a10 = a[4],
        a11 = a[5],
        a12 = a[6],
        a13 = a[7];

    if (a !== out) { // If the source and destination differ, copy the unchanged last row
        out[8]  = a[8];
        out[9]  = a[9];
        out[10] = a[10];
        out[11] = a[11];
        out[12] = a[12];
        out[13] = a[13];
        out[14] = a[14];
        out[15] = a[15];
    }

    // Perform axis-specific matrix multiplication
    out[0] = a00 * c + a10 * s;
    out[1] = a01 * c + a11 * s;
    out[2] = a02 * c + a12 * s;
    out[3] = a03 * c + a13 * s;
    out[4] = a10 * c - a00 * s;
    out[5] = a11 * c - a01 * s;
    out[6] = a12 * c - a02 * s;
    out[7] = a13 * c - a03 * s;
    return out;
};

/**
 * Creates a matrix from a vector translation
 * This is equivalent to (but much faster than):
 *
 *     mat4.identity(dest);
 *     mat4.translate(dest, dest, vec);
 *
 * @param {mat4} out mat4 receiving operation result
 * @param {vec3} v Translation vector
 * @returns {mat4} out
 */
mat4.fromTranslation = function(out, v) {
    out[0] = 1;
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
    out[4] = 0;
    out[5] = 1;
    out[6] = 0;
    out[7] = 0;
    out[8] = 0;
    out[9] = 0;
    out[10] = 1;
    out[11] = 0;
    out[12] = v[0];
    out[13] = v[1];
    out[14] = v[2];
    out[15] = 1;
    return out;
}

/**
 * Creates a matrix from a vector scaling
 * This is equivalent to (but much faster than):
 *
 *     mat4.identity(dest);
 *     mat4.scale(dest, dest, vec);
 *
 * @param {mat4} out mat4 receiving operation result
 * @param {vec3} v Scaling vector
 * @returns {mat4} out
 */
mat4.fromScaling = function(out, v) {
    out[0] = v[0];
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
    out[4] = 0;
    out[5] = v[1];
    out[6] = 0;
    out[7] = 0;
    out[8] = 0;
    out[9] = 0;
    out[10] = v[2];
    out[11] = 0;
    out[12] = 0;
    out[13] = 0;
    out[14] = 0;
    out[15] = 1;
    return out;
}

/**
 * Creates a matrix from a given angle around a given axis
 * This is equivalent to (but much faster than):
 *
 *     mat4.identity(dest);
 *     mat4.rotate(dest, dest, rad, axis);
 *
 * @param {mat4} out mat4 receiving operation result
 * @param {Number} rad the angle to rotate the matrix by
 * @param {vec3} axis the axis to rotate around
 * @returns {mat4} out
 */
mat4.fromRotation = function(out, rad, axis) {
    var x = axis[0], y = axis[1], z = axis[2],
        len = Math.sqrt(x * x + y * y + z * z),
        s, c, t;
    
    if (Math.abs(len) < GLMAT_EPSILON) { return null; }
    
    len = 1 / len;
    x *= len;
    y *= len;
    z *= len;
    
    s = Math.sin(rad);
    c = Math.cos(rad);
    t = 1 - c;
    
    // Perform rotation-specific matrix multiplication
    out[0] = x * x * t + c;
    out[1] = y * x * t + z * s;
    out[2] = z * x * t - y * s;
    out[3] = 0;
    out[4] = x * y * t - z * s;
    out[5] = y * y * t + c;
    out[6] = z * y * t + x * s;
    out[7] = 0;
    out[8] = x * z * t + y * s;
    out[9] = y * z * t - x * s;
    out[10] = z * z * t + c;
    out[11] = 0;
    out[12] = 0;
    out[13] = 0;
    out[14] = 0;
    out[15] = 1;
    return out;
}

/**
 * Creates a matrix from the given angle around the X axis
 * This is equivalent to (but much faster than):
 *
 *     mat4.identity(dest);
 *     mat4.rotateX(dest, dest, rad);
 *
 * @param {mat4} out mat4 receiving operation result
 * @param {Number} rad the angle to rotate the matrix by
 * @returns {mat4} out
 */
mat4.fromXRotation = function(out, rad) {
    var s = Math.sin(rad),
        c = Math.cos(rad);
    
    // Perform axis-specific matrix multiplication
    out[0]  = 1;
    out[1]  = 0;
    out[2]  = 0;
    out[3]  = 0;
    out[4] = 0;
    out[5] = c;
    out[6] = s;
    out[7] = 0;
    out[8] = 0;
    out[9] = -s;
    out[10] = c;
    out[11] = 0;
    out[12] = 0;
    out[13] = 0;
    out[14] = 0;
    out[15] = 1;
    return out;
}

/**
 * Creates a matrix from the given angle around the Y axis
 * This is equivalent to (but much faster than):
 *
 *     mat4.identity(dest);
 *     mat4.rotateY(dest, dest, rad);
 *
 * @param {mat4} out mat4 receiving operation result
 * @param {Number} rad the angle to rotate the matrix by
 * @returns {mat4} out
 */
mat4.fromYRotation = function(out, rad) {
    var s = Math.sin(rad),
        c = Math.cos(rad);
    
    // Perform axis-specific matrix multiplication
    out[0]  = c;
    out[1]  = 0;
    out[2]  = -s;
    out[3]  = 0;
    out[4] = 0;
    out[5] = 1;
    out[6] = 0;
    out[7] = 0;
    out[8] = s;
    out[9] = 0;
    out[10] = c;
    out[11] = 0;
    out[12] = 0;
    out[13] = 0;
    out[14] = 0;
    out[15] = 1;
    return out;
}

/**
 * Creates a matrix from the given angle around the Z axis
 * This is equivalent to (but much faster than):
 *
 *     mat4.identity(dest);
 *     mat4.rotateZ(dest, dest, rad);
 *
 * @param {mat4} out mat4 receiving operation result
 * @param {Number} rad the angle to rotate the matrix by
 * @returns {mat4} out
 */
mat4.fromZRotation = function(out, rad) {
    var s = Math.sin(rad),
        c = Math.cos(rad);
    
    // Perform axis-specific matrix multiplication
    out[0]  = c;
    out[1]  = s;
    out[2]  = 0;
    out[3]  = 0;
    out[4] = -s;
    out[5] = c;
    out[6] = 0;
    out[7] = 0;
    out[8] = 0;
    out[9] = 0;
    out[10] = 1;
    out[11] = 0;
    out[12] = 0;
    out[13] = 0;
    out[14] = 0;
    out[15] = 1;
    return out;
}

/**
 * Creates a matrix from a quaternion rotation and vector translation
 * This is equivalent to (but much faster than):
 *
 *     mat4.identity(dest);
 *     mat4.translate(dest, vec);
 *     var quatMat = mat4.create();
 *     quat4.toMat4(quat, quatMat);
 *     mat4.multiply(dest, quatMat);
 *
 * @param {mat4} out mat4 receiving operation result
 * @param {quat4} q Rotation quaternion
 * @param {vec3} v Translation vector
 * @returns {mat4} out
 */
mat4.fromRotationTranslation = function (out, q, v) {
    // Quaternion math
    var x = q[0], y = q[1], z = q[2], w = q[3],
        x2 = x + x,
        y2 = y + y,
        z2 = z + z,

        xx = x * x2,
        xy = x * y2,
        xz = x * z2,
        yy = y * y2,
        yz = y * z2,
        zz = z * z2,
        wx = w * x2,
        wy = w * y2,
        wz = w * z2;

    out[0] = 1 - (yy + zz);
    out[1] = xy + wz;
    out[2] = xz - wy;
    out[3] = 0;
    out[4] = xy - wz;
    out[5] = 1 - (xx + zz);
    out[6] = yz + wx;
    out[7] = 0;
    out[8] = xz + wy;
    out[9] = yz - wx;
    out[10] = 1 - (xx + yy);
    out[11] = 0;
    out[12] = v[0];
    out[13] = v[1];
    out[14] = v[2];
    out[15] = 1;
    
    return out;
};

/**
 * Creates a matrix from a quaternion rotation, vector translation and vector scale
 * This is equivalent to (but much faster than):
 *
 *     mat4.identity(dest);
 *     mat4.translate(dest, vec);
 *     var quatMat = mat4.create();
 *     quat4.toMat4(quat, quatMat);
 *     mat4.multiply(dest, quatMat);
 *     mat4.scale(dest, scale)
 *
 * @param {mat4} out mat4 receiving operation result
 * @param {quat4} q Rotation quaternion
 * @param {vec3} v Translation vector
 * @param {vec3} s Scaling vector
 * @returns {mat4} out
 */
mat4.fromRotationTranslationScale = function (out, q, v, s) {
    // Quaternion math
    var x = q[0], y = q[1], z = q[2], w = q[3],
        x2 = x + x,
        y2 = y + y,
        z2 = z + z,

        xx = x * x2,
        xy = x * y2,
        xz = x * z2,
        yy = y * y2,
        yz = y * z2,
        zz = z * z2,
        wx = w * x2,
        wy = w * y2,
        wz = w * z2,
        sx = s[0],
        sy = s[1],
        sz = s[2];

    out[0] = (1 - (yy + zz)) * sx;
    out[1] = (xy + wz) * sx;
    out[2] = (xz - wy) * sx;
    out[3] = 0;
    out[4] = (xy - wz) * sy;
    out[5] = (1 - (xx + zz)) * sy;
    out[6] = (yz + wx) * sy;
    out[7] = 0;
    out[8] = (xz + wy) * sz;
    out[9] = (yz - wx) * sz;
    out[10] = (1 - (xx + yy)) * sz;
    out[11] = 0;
    out[12] = v[0];
    out[13] = v[1];
    out[14] = v[2];
    out[15] = 1;
    
    return out;
};

/**
 * Creates a matrix from a quaternion rotation, vector translation and vector scale, rotating and scaling around the given origin
 * This is equivalent to (but much faster than):
 *
 *     mat4.identity(dest);
 *     mat4.translate(dest, vec);
 *     mat4.translate(dest, origin);
 *     var quatMat = mat4.create();
 *     quat4.toMat4(quat, quatMat);
 *     mat4.multiply(dest, quatMat);
 *     mat4.scale(dest, scale)
 *     mat4.translate(dest, negativeOrigin);
 *
 * @param {mat4} out mat4 receiving operation result
 * @param {quat4} q Rotation quaternion
 * @param {vec3} v Translation vector
 * @param {vec3} s Scaling vector
 * @param {vec3} o The origin vector around which to scale and rotate
 * @returns {mat4} out
 */
mat4.fromRotationTranslationScaleOrigin = function (out, q, v, s, o) {
  // Quaternion math
  var x = q[0], y = q[1], z = q[2], w = q[3],
      x2 = x + x,
      y2 = y + y,
      z2 = z + z,

      xx = x * x2,
      xy = x * y2,
      xz = x * z2,
      yy = y * y2,
      yz = y * z2,
      zz = z * z2,
      wx = w * x2,
      wy = w * y2,
      wz = w * z2,
      
      sx = s[0],
      sy = s[1],
      sz = s[2],

      ox = o[0],
      oy = o[1],
      oz = o[2];
      
  out[0] = (1 - (yy + zz)) * sx;
  out[1] = (xy + wz) * sx;
  out[2] = (xz - wy) * sx;
  out[3] = 0;
  out[4] = (xy - wz) * sy;
  out[5] = (1 - (xx + zz)) * sy;
  out[6] = (yz + wx) * sy;
  out[7] = 0;
  out[8] = (xz + wy) * sz;
  out[9] = (yz - wx) * sz;
  out[10] = (1 - (xx + yy)) * sz;
  out[11] = 0;
  out[12] = v[0] + ox - (out[0] * ox + out[4] * oy + out[8] * oz);
  out[13] = v[1] + oy - (out[1] * ox + out[5] * oy + out[9] * oz);
  out[14] = v[2] + oz - (out[2] * ox + out[6] * oy + out[10] * oz);
  out[15] = 1;
        
  return out;
};

mat4.fromQuat = function (out, q) {
    var x = q[0], y = q[1], z = q[2], w = q[3],
        x2 = x + x,
        y2 = y + y,
        z2 = z + z,

        xx = x * x2,
        yx = y * x2,
        yy = y * y2,
        zx = z * x2,
        zy = z * y2,
        zz = z * z2,
        wx = w * x2,
        wy = w * y2,
        wz = w * z2;

    out[0] = 1 - yy - zz;
    out[1] = yx + wz;
    out[2] = zx - wy;
    out[3] = 0;

    out[4] = yx - wz;
    out[5] = 1 - xx - zz;
    out[6] = zy + wx;
    out[7] = 0;

    out[8] = zx + wy;
    out[9] = zy - wx;
    out[10] = 1 - xx - yy;
    out[11] = 0;

    out[12] = 0;
    out[13] = 0;
    out[14] = 0;
    out[15] = 1;

    return out;
};

/**
 * Generates a frustum matrix with the given bounds
 *
 * @param {mat4} out mat4 frustum matrix will be written into
 * @param {Number} left Left bound of the frustum
 * @param {Number} right Right bound of the frustum
 * @param {Number} bottom Bottom bound of the frustum
 * @param {Number} top Top bound of the frustum
 * @param {Number} near Near bound of the frustum
 * @param {Number} far Far bound of the frustum
 * @returns {mat4} out
 */
mat4.frustum = function (out, left, right, bottom, top, near, far) {
    var rl = 1 / (right - left),
        tb = 1 / (top - bottom),
        nf = 1 / (near - far);
    out[0] = (near * 2) * rl;
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
    out[4] = 0;
    out[5] = (near * 2) * tb;
    out[6] = 0;
    out[7] = 0;
    out[8] = (right + left) * rl;
    out[9] = (top + bottom) * tb;
    out[10] = (far + near) * nf;
    out[11] = -1;
    out[12] = 0;
    out[13] = 0;
    out[14] = (far * near * 2) * nf;
    out[15] = 0;
    return out;
};

/**
 * Generates a perspective projection matrix with the given bounds
 *
 * @param {mat4} out mat4 frustum matrix will be written into
 * @param {number} fovy Vertical field of view in radians
 * @param {number} aspect Aspect ratio. typically viewport width/height
 * @param {number} near Near bound of the frustum
 * @param {number} far Far bound of the frustum
 * @returns {mat4} out
 */
mat4.perspective = function (out, fovy, aspect, near, far) {
    var f = 1.0 / Math.tan(fovy / 2),
        nf = 1 / (near - far);
    out[0] = f / aspect;
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
    out[4] = 0;
    out[5] = f;
    out[6] = 0;
    out[7] = 0;
    out[8] = 0;
    out[9] = 0;
    out[10] = (far + near) * nf;
    out[11] = -1;
    out[12] = 0;
    out[13] = 0;
    out[14] = (2 * far * near) * nf;
    out[15] = 0;
    return out;
};

/**
 * Generates a perspective projection matrix with the given field of view.
 * This is primarily useful for generating projection matrices to be used
 * with the still experiemental WebVR API.
 *
 * @param {mat4} out mat4 frustum matrix will be written into
 * @param {number} fov Object containing the following values: upDegrees, downDegrees, leftDegrees, rightDegrees
 * @param {number} near Near bound of the frustum
 * @param {number} far Far bound of the frustum
 * @returns {mat4} out
 */
mat4.perspectiveFromFieldOfView = function (out, fov, near, far) {
    var upTan = Math.tan(fov.upDegrees * Math.PI/180.0),
        downTan = Math.tan(fov.downDegrees * Math.PI/180.0),
        leftTan = Math.tan(fov.leftDegrees * Math.PI/180.0),
        rightTan = Math.tan(fov.rightDegrees * Math.PI/180.0),
        xScale = 2.0 / (leftTan + rightTan),
        yScale = 2.0 / (upTan + downTan);

    out[0] = xScale;
    out[1] = 0.0;
    out[2] = 0.0;
    out[3] = 0.0;
    out[4] = 0.0;
    out[5] = yScale;
    out[6] = 0.0;
    out[7] = 0.0;
    out[8] = -((leftTan - rightTan) * xScale * 0.5);
    out[9] = ((upTan - downTan) * yScale * 0.5);
    out[10] = far / (near - far);
    out[11] = -1.0;
    out[12] = 0.0;
    out[13] = 0.0;
    out[14] = (far * near) / (near - far);
    out[15] = 0.0;
    return out;
}

/**
 * Generates a orthogonal projection matrix with the given bounds
 *
 * @param {mat4} out mat4 frustum matrix will be written into
 * @param {number} left Left bound of the frustum
 * @param {number} right Right bound of the frustum
 * @param {number} bottom Bottom bound of the frustum
 * @param {number} top Top bound of the frustum
 * @param {number} near Near bound of the frustum
 * @param {number} far Far bound of the frustum
 * @returns {mat4} out
 */
mat4.ortho = function (out, left, right, bottom, top, near, far) {
    var lr = 1 / (left - right),
        bt = 1 / (bottom - top),
        nf = 1 / (near - far);
    out[0] = -2 * lr;
    out[1] = 0;
    out[2] = 0;
    out[3] = 0;
    out[4] = 0;
    out[5] = -2 * bt;
    out[6] = 0;
    out[7] = 0;
    out[8] = 0;
    out[9] = 0;
    out[10] = 2 * nf;
    out[11] = 0;
    out[12] = (left + right) * lr;
    out[13] = (top + bottom) * bt;
    out[14] = (far + near) * nf;
    out[15] = 1;
    return out;
};

/**
 * Generates a look-at matrix with the given eye position, focal point, and up axis
 *
 * @param {mat4} out mat4 frustum matrix will be written into
 * @param {vec3} eye Position of the viewer
 * @param {vec3} center Point the viewer is looking at
 * @param {vec3} up vec3 pointing up
 * @returns {mat4} out
 */
mat4.lookAt = function (out, eye, center, up) {
    var x0, x1, x2, y0, y1, y2, z0, z1, z2, len,
        eyex = eye[0],
        eyey = eye[1],
        eyez = eye[2],
        upx = up[0],
        upy = up[1],
        upz = up[2],
        centerx = center[0],
        centery = center[1],
        centerz = center[2];

    if (Math.abs(eyex - centerx) < GLMAT_EPSILON &&
        Math.abs(eyey - centery) < GLMAT_EPSILON &&
        Math.abs(eyez - centerz) < GLMAT_EPSILON) {
        return mat4.identity(out);
    }

    z0 = eyex - centerx;
    z1 = eyey - centery;
    z2 = eyez - centerz;

    len = 1 / Math.sqrt(z0 * z0 + z1 * z1 + z2 * z2);
    z0 *= len;
    z1 *= len;
    z2 *= len;

    x0 = upy * z2 - upz * z1;
    x1 = upz * z0 - upx * z2;
    x2 = upx * z1 - upy * z0;
    len = Math.sqrt(x0 * x0 + x1 * x1 + x2 * x2);
    if (!len) {
        x0 = 0;
        x1 = 0;
        x2 = 0;
    } else {
        len = 1 / len;
        x0 *= len;
        x1 *= len;
        x2 *= len;
    }

    y0 = z1 * x2 - z2 * x1;
    y1 = z2 * x0 - z0 * x2;
    y2 = z0 * x1 - z1 * x0;

    len = Math.sqrt(y0 * y0 + y1 * y1 + y2 * y2);
    if (!len) {
        y0 = 0;
        y1 = 0;
        y2 = 0;
    } else {
        len = 1 / len;
        y0 *= len;
        y1 *= len;
        y2 *= len;
    }

    out[0] = x0;
    out[1] = y0;
    out[2] = z0;
    out[3] = 0;
    out[4] = x1;
    out[5] = y1;
    out[6] = z1;
    out[7] = 0;
    out[8] = x2;
    out[9] = y2;
    out[10] = z2;
    out[11] = 0;
    out[12] = -(x0 * eyex + x1 * eyey + x2 * eyez);
    out[13] = -(y0 * eyex + y1 * eyey + y2 * eyez);
    out[14] = -(z0 * eyex + z1 * eyey + z2 * eyez);
    out[15] = 1;

    return out;
};

/**
 * Returns a string representation of a mat4
 *
 * @param {mat4} mat matrix to represent as a string
 * @returns {String} string representation of the matrix
 */
mat4.str = function (a) {
    return 'mat4(' + a[0] + ', ' + a[1] + ', ' + a[2] + ', ' + a[3] + ', ' +
                    a[4] + ', ' + a[5] + ', ' + a[6] + ', ' + a[7] + ', ' +
                    a[8] + ', ' + a[9] + ', ' + a[10] + ', ' + a[11] + ', ' + 
                    a[12] + ', ' + a[13] + ', ' + a[14] + ', ' + a[15] + ')';
};

/**
 * Returns Frobenius norm of a mat4
 *
 * @param {mat4} a the matrix to calculate Frobenius norm of
 * @returns {Number} Frobenius norm
 */
mat4.frob = function (a) {
    return(Math.sqrt(Math.pow(a[0], 2) + Math.pow(a[1], 2) + Math.pow(a[2], 2) + Math.pow(a[3], 2) + Math.pow(a[4], 2) + Math.pow(a[5], 2) + Math.pow(a[6], 2) + Math.pow(a[7], 2) + Math.pow(a[8], 2) + Math.pow(a[9], 2) + Math.pow(a[10], 2) + Math.pow(a[11], 2) + Math.pow(a[12], 2) + Math.pow(a[13], 2) + Math.pow(a[14], 2) + Math.pow(a[15], 2) ))
};


if(typeof(exports) !== 'undefined') {
    exports.mat4 = mat4;
}
var gl, ctx;
var tempCanvas, globalCanvas;
var projectionMatrix;

var UserUpdate, UserDraw;

function EngineMain(canvas, loadCB, updateCB, drawCB) {
    globalCanvas = canvas;
    InitGL(canvas);
    InitTempCanvas();
    
    InitSpriteParams();
    SetupTextRendering();
	SetupInput();
    
    if(loadCB) loadCB();
    
    projectionMatrix = mat4.create();
    mat4.ortho(projectionMatrix, 0, canvas.width, canvas.height, 0, -1000, 1000);
    
    if(!updateCB) {
        console.error("You need to specify an update function!");
        return;
    }
    
    if(!drawCB) {
        console.error("You need to specify a draw function!");
        return;
    }
    
    UserUpdate = updateCB;
    UserDraw = drawCB;
    
    var loop = setInterval(Loop, 15);
}

function InitGL(canvas) {
    try {
        gl = canvas.getContext("experimental-webgl");
        gl.viewportWidth = canvas.width;        
        gl.viewportHeight = canvas.height;
        gl.enable(gl.BLEND)
        gl.blendFunc(gl.SRC_ALPHA, gl.ONE);
        gl.clearColor(1, 0, 0, 1);
    } catch(e) {
    }
    
    if(!gl) {
        alert("WebGL failed to load.");    
    }    
}

function InitTempCanvas() {
    tempCanvas = document.createElement('canvas');
    tempCanvas.id = "TempCanvas";
    tempCanvas.width = 1000;
    tempCanvas.height = 1000;
    document.body.appendChild(tempCanvas);
    tempCanvas.style.visibility = "hidden";
    ctx = tempCanvas.getContext("2d");
    
    if(!tempCanvas) console.log("Temp canvas failed!"); 
    if(!ctx) console.log("2D context failed!");
}

function Loop() {
    Update((1 / 1000) * 16.666667);    
    Draw((1 / 1000) * 16.666667);
}

function Update(dt) {
    UserUpdate(dt);
}

function Draw(dt) {
    gl.clear(gl.COLOR_BUFFER_BIT | gl.DEPTH_BUFFER_BIT);
    UserDraw(dt);
}function SetupInput() {
    window.onkeydown = function(e) {
        return !(e.keyCode == 32);
    }
    
    document.addEventListener('keydown', function(event) {Keyboard.OnKeydown(event);});
    document.addEventListener('keyup', function(event) {Keyboard.OnKeyup(event);});
    document.body.onmousedown = function() { 
        ++mouseDown;
        mouseClick = 1;
    }
    document.body.onmouseup = function() {
        --mouseDown;
    }
    
    globalCanvas.addEventListener('mousemove', function(e){ 
        var rect = globalCanvas.getBoundingClientRect();
        mouse.x = e.clientX - rect.left; 
        mouse.y = e.clientY - rect.top;
    }, false);
}

var Keyboard = {
    _pressed: [],

    IsDown: function(keyCode) {
        return this._pressed[keyCode];
    },

    OnKeydown: function(event) {
        this._pressed[event.keyCode] = true;
    },

    OnKeyup: function(event) {
        delete this._pressed[event.keyCode];
    }
};

var Keys = {
    ENTER: 13,
    SHIFT: 16,
    CTRL: 17,
    ALT: 18,
    ESCAPE: 27,
    SPACE: 32,
    
    LEFT: 37,
    UP: 38,
    RIGHT: 39,
    DOWN: 40,
    
    ZERO: 48,
    ONE: 49,
    TWO: 50,
    THREE: 51,
    FOUR: 52,
    FIVE: 53,
    SIX: 54,
    SEVEN: 55,
    EIGHT: 56,
    NINE: 57,
    
    A: 65,
    B: 66,
    C: 67,
    D: 68,
    E: 69,
    F: 70,
    G: 71,
    H: 72,
    I: 73,
    J: 74,
    K: 75,
    L: 76,
    M: 77,
    M: 78,
    O: 79,
    P: 80,
    Q: 81,
    R: 82,
    S: 83,
    T: 84,
    U: 85,
    V: 86,
    W: 87,
    X: 88,
    Y: 89,
    Z: 90,
    
    NUM_ZERO: 96,
    NUM_ONE: 97,
    NUM_TWO: 98,
    NUM_THREE: 99,
    NUM_FOUR: 100,
    NUM_FIVE: 101,
    NUM_SIX: 102,
    NUM_SEVEN: 103,
    NUM_EIGHT: 104,
    NUM_NINE: 105
}

var mouseDown = 0;
var mouseClick = 0;
var mouse = {x: 0, y: 0};Parent = {};  

Parent.Extend = function (subClass, baseClass) {
	function inheritance() { }
	inheritance.prototype = baseClass.prototype;
	subClass.prototype = new inheritance();
	subClass.prototype.constructor = subClass;
	subClass.base = baseClass;
	subClass.superClass = baseClass.prototype;
}
var BASIC_SHADER_FS = MultiLineString(function() {/*
precision mediump float;
uniform vec4 colour;
uniform sampler2D texture;

varying vec2 coords;

void main() {
    gl_FragColor = texture2D(texture, coords) + colour;
}
*/
});
      
var BASIC_SHADER_VS = MultiLineString(function() {/*
precision mediump float;
attribute vec3 position;
attribute vec2 texCoord;

varying vec2 coords;

uniform mat4 model;
uniform mat4 projection;

void main(void) {
    coords = texCoord;
    gl_Position = projection * model * vec4(position, 1.0);
}
*/
});

var SIMPLE_SHADER;

function Shader(vs, fs) {
    this.id = gl.createProgram();
    
    var vertex = this.AddShader(vs, gl.VERTEX_SHADER);
    var fragment = this.AddShader(fs, gl.FRAGMENT_SHADER);
    
    gl.attachShader(this.id, vertex);
    gl.attachShader(this.id, fragment);
    gl.linkProgram(this.id);
    
    if (!gl.getProgramParameter(this.id, gl.LINK_STATUS)) {
      alert("Could not initialise shaders");
    }

    gl.useProgram(this.id);
    
    this.positionAttribute = gl.getAttribLocation(this.id, "position");
    this.texCoordAttribute = gl.getAttribLocation(this.id, "texCoord");
    this.projectionUniform = gl.getUniformLocation(this.id, "projection");
    this.modelviewUniform = gl.getUniformLocation(this.id, "model");
    this.colourUniform = gl.getUniformLocation(this.id, "colour");
    this.textureUniform = gl.getUniformLocation(this.id, "texture");
}

Shader.prototype.AddShader = function(src, type) {
    var shader = gl.createShader(type);
    gl.shaderSource(shader, src);
    gl.compileShader(shader);
    
    if (!gl.getShaderParameter(shader, gl.COMPILE_STATUS)) {
          console.error(gl.getShaderInfoLog(shader));
          return null;
    }

    return shader;
}

Shader.prototype.Bind = function() {
    gl.useProgram(this.id);
}

Shader.prototype.UniformMatrix = function(pos, val) {
    gl.uniformMatrix4fv(pos, false, val);
}

Shader.prototype.UniformFloat4 = function(pos, val) {
    gl.uniform4f(pos, val[0], val[1], val[2], val[3]);
}


Shader.prototype.UniformTexture = function(pos, val) {
    gl.uniform1i(pos, val);
}
function Sound(name) {
    this.audio = new Audio(name);
}

Sound.prototype.Play = function() {
    this.audio.loop = false;
    this.audio.play();
}

Sound.prototype.Pause = function() {
    this.audio.pause();
}

Sound.prototype.Stop = function() {
    this.audio.pause();
    this.audio.currentTime = 0;
}

Sound.prototype.Loop = function() {
    this.audio.loop = true;
    this.audio.play();
    this.audio.addEventListener('timeupdate', function(){
                var buffer = .30
                if(this.currentTime > this.duration - buffer){
                    this.currentTime = 0
                    this.play()
                }}, false);
}

Sound.prototype.SetVolume = function(volume) {
    this.audio.volume = volume;
}
var SPRITE_VERTICES = [
    0.5, 0.5, 0, 1, 1,
    -0.5, 0.5, 0, 0, 1,
    0.5, -0.5, 0, 1, 0,
    -0.5, -0.5, 0, 0, 0
];

var SPRITE_BUFFER;
var SPRITE_CURRENT_SHADER;

function InitSpriteParams() {
    SPRITE_BUFFER = gl.createBuffer();
    gl.bindBuffer(gl.ARRAY_BUFFER, SPRITE_BUFFER);
    gl.bufferData(gl.ARRAY_BUFFER, new Float32Array(SPRITE_VERTICES), gl.STATIC_DRAW);
    SPRITE_BUFFER.vertexSize = 3;
    SPRITE_BUFFER.texSize = 2;
    SPRITE_BUFFER.numItems = 4;
}

function Sprite(texture, color) {
    if(!texture) {
        this.texture = null;
    } else {
        this.texture = texture;
    }
    
    if(!color) color = [0, 0, 0, 0];
    this.color = color;
}

var SpriteSetShader = function(shader) {
    SPRITE_CURRENT_SHADER = shader;
    gl.bindBuffer(gl.ARRAY_BUFFER, SPRITE_BUFFER);
    
    SPRITE_CURRENT_SHADER.Bind();
    gl.enableVertexAttribArray(SPRITE_CURRENT_SHADER.positionAttribute);
    gl.enableVertexAttribArray(SPRITE_CURRENT_SHADER.texCoordAttribute);
    gl.vertexAttribPointer(SPRITE_CURRENT_SHADER.positionAttribute, SPRITE_BUFFER.vertexSize, gl.FLOAT, false, 20, 0);            
    gl.vertexAttribPointer(SPRITE_CURRENT_SHADER.texCoordAttribute, SPRITE_BUFFER.texSize, gl.FLOAT, false, 20, 12);

    SPRITE_CURRENT_SHADER.UniformMatrix(SPRITE_CURRENT_SHADER.projectionUniform, projectionMatrix);
}

Sprite.prototype.Draw = function(x, y, scaleX, scaleY, rot) {
    if(!scaleY) scaleY = scaleX;
    
    var modelMat = mat4.create();
    mat4.translate(modelMat, modelMat, [x, y, 0]);
    mat4.scale(modelMat, modelMat, [scaleX, scaleY, 1]);
    
    if(rot) {
        mat4.rotate(modelMat, modelMat, ToRadian(rot), [0, 0, 1]);
    }
    
    if(this.texture) {
        gl.activeTexture(gl.TEXTURE0);
        this.texture.Bind();
        SPRITE_CURRENT_SHADER.UniformTexture(SPRITE_CURRENT_SHADER.textureUniform, 0);
    }
    
    SPRITE_CURRENT_SHADER.UniformMatrix(SPRITE_CURRENT_SHADER.modelviewUniform, modelMat);
    
    if(SPRITE_CURRENT_SHADER.colourUniform != -1) {
        SPRITE_CURRENT_SHADER.UniformFloat4(SPRITE_CURRENT_SHADER.colourUniform, this.color);
    }
    
    gl.drawArrays(gl.TRIANGLE_STRIP, 0, SPRITE_BUFFER.numItems);
    
    if(this.texture) this.texture.Unbind();
    
}
function SetupTextRendering() {
    ctx.font = "12px monospace";
}

function Text(text, fontSize, font) {
    this.text = text;
    this.fontSize = fontSize;
    this.texture = null;
    this.sprite = null;
    this.width = 0;
    this.height = 0;
    this.font = font || "monospace";
    this.GetTexture();
}

Text.prototype.GetTexture = function() {    
    ctx.font = this.fontSize+"px "+this.font;
    var temp = ctx.measureText(this.text);
    tempCanvas.width = temp.width;
    tempCanvas.height = this.fontSize;
    
    this.width = tempCanvas.width;
    this.height = tempCanvas.height;
    
    ctx.clearRect(0, 0, tempCanvas.width, tempCanvas.height);
    ctx.fillStyle = "#ffffff";
    ctx.textAlign = "center";
    ctx.textBaseline = "middle";
    ctx.font = this.fontSize+"px monospace";

    ctx.fillText(this.text, tempCanvas.width/2, tempCanvas.height/2);

    this.texture = new Texture("TempCanvas", true);
    this.sprite = new Sprite(this.texture);
}

Text.prototype.SetText = function(newText, fontSize) {
    this.fontSize = fontSize || this.fontSize;
    this.text = newText;
    this.GetTexture();
}

Text.prototype.Draw = function(x, y) {
    this.sprite.Draw(x + this.width / 2, y + this.height / 2, this.width, this.height);
}

var TEMP_TEXTURE = false;

var imgs = {};
var texs = {};

function Texture(src, isOnDoc) {
    if(!TEMP_TEXTURE) {
        TEMP_TEXTURE = gl.createTexture();
        gl.bindTexture(gl.TEXTURE_2D, TEMP_TEXTURE);
        gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, 1, 1, 0, gl.RGBA, gl.UNSIGNED_BYTE, new Uint8Array([255, 0, 0, 255])); // red
    }
    
    this.id = TEMP_TEXTURE
    
    if(isOnDoc) {
        var img = document.getElementById(src);
        texs[img.src] = this;
        texs[img.src].id = gl.createTexture();
        gl.bindTexture(gl.TEXTURE_2D, texs[img.src].id);
        gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, img);
        texs[img.src].Setup(img.width, img.height);
    } else {
        var img = new Image();
        img.crossOrigin = '';
        img.src = src;
        texs[img.src] = this;
        img.onload = function() {
            texs[img.src].id = gl.createTexture();
            gl.bindTexture(gl.TEXTURE_2D, texs[img.src].id);
            gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, img);
            texs[img.src].Setup(img.width, img.height);
        }
    }
}

Texture.prototype.Setup = function(width, height) {
    if (IsPowerOf2(width) && IsPowerOf2(height)) {
        gl.generateMipmap(gl.TEXTURE_2D);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
    } else {
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_S, gl.CLAMP_TO_EDGE);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_WRAP_T, gl.CLAMP_TO_EDGE);
        gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
    }
    
    gl.bindTexture(gl.TEXTURE_2D, null);
}

function TextureCallbackLoad(texture, data) {
    texture.Load(data);
}

Texture.prototype.Load = function(data) {
    gl.bindTexture(gl.TEXTURE_2D, this.id);
    gl.texImage2D(gl.TEXTURE_2D, 0, gl.RGBA, gl.RGBA, gl.UNSIGNED_BYTE, data);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MAG_FILTER, gl.NEAREST);
    gl.texParameteri(gl.TEXTURE_2D, gl.TEXTURE_MIN_FILTER, gl.NEAREST);
    gl.bindTexture(gl.TEXTURE_2D, null);
}

Texture.prototype.Bind = function() {
       gl.bindTexture(gl.TEXTURE_2D, this.id);
}

Texture.prototype.Unbind = function() {
   gl.bindTexture(gl.TEXTURE_2D, null);
}
var DEG_TO_RAD = Math.PI / 180.0;

function MultiLineString(f) {
  return f.toString().
      replace(/^[^\/]+\/\*!?/, '').
      replace(/\*\/[^\/]+$/, '');
}


function ToRadian(deg) {
    return deg * DEG_TO_RAD;
}

function IsPowerOf2(value) {
    return (value & (value - 1)) == 0;
};

function GetPowerOfTwo(value, pow) {
    var pow = pow || 1;
    
    while(pow<value) {
        pow *= 2;
    }
    
    return pow;
}
window.addEventListener('load', EngineMain(document.getElementById("canvas"), Load, Update, Draw), false);

var TOP_SCORE = GetCookie("TopScore");

var frames, fps, timer;
var text, scoreText, currentScoreText;

var score = 0;

var shakeX = 0, shakeY = 0;
var hasStarted = false;
var hasDied = false;

var blockSound = new Sound("blockhit.wav");
var enemySound = new Sound("enemyhit.wav");
var playerSound = new Sound("playerhit.wav");
var music = new Sound("music.ogg");
music.audio.addEventListener('canplay', function() {
        if(!isPlaying) {
        music.SetVolume(0.25);
        music.Loop();
        isPlaying = true;
        playerSound.SetVolume(0.5);
    }
});
var isPlaying = false;

var playerParticles = [];
var playerParticleSprite = new Sprite(null, [1.0, 0.0, 1.0, 1.0]);

var enemyParticles = [];
var enemyParticleSprite = new Sprite(null, [1.0, 0.2, 0.2, 1.0]);

var muteMusicTex = new Sprite(new Texture("MusicTex.png"));
var unmuteMusicTex = new Sprite(new Texture("MusicTexMuted.png"));
var isMusicMuted = false;
var musicRectangle = {
    left: 800-32,
    right: 800,
    top: 600-10-64,
    bottom: 600-10-32
};

var muteSoundTex = new Sprite(new Texture("SoundTex.png"));
var unmuteSoundTex = new Sprite(new Texture("SoundTexMuted.png"));
var isSoundMuted = false;
var soundRectangle = {
    left: 800-32,
    right: 800,
    top: 600-5-32,
    bottom: 600-5
};

var comboText;
var comboTimer = 0, comboMaxTimer = 0.5;
var combo = 0;

var comboStrings = [
    "Double kill",
    "Triple kill!",
    "Quadra kill!!",
    "PENTA KILL!!!",
    "Too much! Calm down!",
];

var player = {
    x: 400,
    y: 300,
    size: 32,
    
    isPushing: false,    
    outwards: false,
    pushTimer: 0,
    pushMaxTimer: 0.5,
    pushSize: 0,
    pushMaxSize: 76,
    
    isPulling: false,
    pullSize: 0,
    pullMaxSize: 80,
    pullTimer: 0,
    pullMaxTimer: 1,
    
    pushSprite: new Sprite(null, [0.6, 0.6, 0.6, 0.6]),
    pullSprite: new Sprite(null, [0.6, 0.6, 0.6, 0.6]),
    sprite: new Sprite(null, [1.0, 0.0, 1.0, 1.0])
};

var block = {
    x: 400,
    y: 150,
    velx: 0,
    vely: 0,
    friction: 0.9,
    size: 24,
    sprite: new Sprite(null, [1.0, 0.7, 0.3, 1.0]),
    motionBlur: []
};

var blockBlurSprites = [
    new Sprite(null, [1.0, 0.7, 0.3, 1.0]),
    new Sprite(null, [1.0, 0.7, 0.3, 0.8]),
    new Sprite(null, [1.0, 0.7, 0.3, 0.6]),
    new Sprite(null, [1.0, 0.7, 0.3, 0.4]),
    new Sprite(null, [1.0, 0.7, 0.3, 0.2]),
];

var enemySprite = new Sprite(null, [1.0, 0.2, 0.2, 1.0]);

var enemies = {
    array: []
};

var particles = [];
var particleSprite = new Sprite(null, [0.8, 0.3, 0.3, 1.0]);

var enemyTimer = 0;

function Load() {
    gl.clearColor(1, 1, 0, 1);
    SIMPLE_SHADER = new Shader(BASIC_SHADER_VS, BASIC_SHADER_FS);
	
    text = new Text("temp", 22);    
    currentScoreText = new Text("temp", 22);
    scoreText = new Text("Score:", 36);
    comboText = new Text("null", 36);

    fps = 0;
    frames = 0;
    timer = Date.now();
}

function Update(dt) {
    var mouseRect = {
        left: mouse.x,
        right: mouse.x + 1,
        top: mouse.y,
        bottom: mouse.y + 1
    };
    
    if(Date.now() - timer >= 1000) {
        fps = frames;
        frames = 0;
        timer = Date.now();
        text.SetText(fps + " fps");
    }
    
    enemyTimer += dt;
    if(enemyTimer > Math.max((4 - score / 15), 0.5)  && hasStarted) {
        CreateEnemy();
        enemyTimer = 0;
    }

    var xLeft = block.velx;
    var yLeft = block.vely;
    
    while(Math.abs(xLeft) > 0 || Math.abs(yLeft) > 0) {
        if(Math.abs(xLeft) >= 1) {
            block.x += Sign(xLeft);
            xLeft -= Sign(xLeft);
        } else {
            block.x += xLeft;
            xLeft = 0;
        }
        
        if(Math.abs(yLeft) >= 1) {
            block.y += Sign(yLeft);
            yLeft -= Sign(yLeft);
        } else {
            block.y += yLeft;
            yLeft = 0;
        }
        
        BlockEnemyCollision();
    }

    block.velx *= block.friction;
    block.vely *= block.friction;
    
    if(!hasDied) {
        if(Keyboard.IsDown(Keys.W)) player.y -= 5;
        if(Keyboard.IsDown(Keys.A)) player.x -= 5;
        if(Keyboard.IsDown(Keys.S)) player.y += 5;
        if(Keyboard.IsDown(Keys.D)) player.x += 5;
        playerParticles.push({
                    x: player.x,
                    y: player.y,
                    life: 5 + Math.random() * 25,
                    velx: (Math.random() - 0.5) * 1.1,
                    vely: (Math.random() - 0.5) * 1.1,
                    size: Math.random() * 12
        });
    }
    
    player.pushTimer += dt;
    player.pullTimer += dt;
    comboTimer += dt;
    
    if(comboTimer >= comboMaxTimer) {
        combo = 0;
    }
    
    if(Keyboard.IsDown(Keys.SHIFT) && player.isPulling === false && player.pullTimer > player.pullMaxTimer) {
        player.isPulling = true;
        player.pullTimer = 0;
        player.pullSize = player.pullMaxSize;
        block.velx = -(block.x - player.x) / 20;
        block.vely = -(block.y - player.y) / 20;
    }
    
    if(Keyboard.IsDown(Keys.SPACE) && player.isPushing === false && player.pushTimer > player.pushMaxTimer) {
        player.isPushing = true;
        player.outwards = true;
        player.pushTimer = 0;
    }
    
    if(player.isPushing) {
        if(player.outwards) {
            player.pushSize += player.pushMaxSize / (player.pushMaxTimer/6) * dt;
            if(player.pushSize >= player.pushMaxSize) {
                player.outwards = false;
            }
        } else {
            player.pushSize -= player.pushMaxSize / (player.pushMaxTimer/6) * dt;
            if(player.pushSize <= 0) {
                player.pushSize = 0;
                player.isPushing = false;
            }
        }
    }
    
    if(player.isPulling) {
        player.pullSize -= player.pullMaxSize / (player.pullMaxTimer/6) * dt;
        if(player.pullSize <= 0) {
            player.pullSize = 0;
            player.isPulling = false;
        }
    }
    
    if(player.x < player.size/2) player.x = player.size/2;
    if(player.x > 800-player.size/2) player.x = 800-player.size/2;
    if(player.y < player.size/2) player.y = player.size/2;
    if(player.y > 600-player.size/2) player.y = 600-player.size/2;
    
    if(block.x < block.size/2) {
        block.x = block.size/2;
        block.velx = -block.velx;
        if(Math.abs(block.velx) > 6) {
            shakeX += 10;
            shakeY += 10;
            blockSound.Stop();
            blockSound.Play();
        }
    }
    
    if(block.x > 800 - block.size/2) {
        block.x = 800 - block.size/2;
        block.velx = -block.velx;
        if(Math.abs(block.velx) > 6) {
            shakeX += 10;
            shakeY += 10;
            blockSound.Stop();
            blockSound.Play();
        }
    }
    
    if(block.y < block.size/2) {
        block.y = block.size/2;
        block.vely = -block.vely;
        if(Math.abs(block.vely) > 6) {
            shakeX += 10;
            shakeY += 10;
            blockSound.Stop();
            blockSound.Play();
        }
    }
    
    if(block.y > 600 - block.size/2) {
        block.y = 600 - block.size/2;
        block.vely = -block.vely;
        if(Math.abs(block.vely) > 6) {
            shakeX += 10;
            shakeY += 10;
            blockSound.Stop();
            blockSound.Play();
        }
    }
    
    if(player.outwards == true && RectangleCollision({
                              left: player.x - player.pushSize / 2,
                              right: player.x + player.pushSize / 2,
                              top: player.y - player.pushSize / 2,
                              bottom: player.y + player.pushSize / 2
                          }, {
                              left: block.x - block.size / 2,
                              right: block.x + block.size / 2,
                              top: block.y - block.size / 2,
                              bottom: block.y + block.size / 2
                        })) {
        block.velx = (block.x - player.x)*2;
        block.vely = (block.y - player.y)*2;
        shakeX += 50;
        shakeY += 50;
        blockSound.Stop();
        blockSound.Play();
        if(!hasStarted) {
            score = 0;
            currentScoreText.SetText("Score: 0");
            hasStarted = true;
        }
    }
    
    if(RectangleCollision({
                              left: player.x - player.size / 2,
                              right: player.x + player.size / 2,
                              top: player.y - player.size / 2,
                              bottom: player.y + player.size / 2
                          }, {
                              left: block.x - block.size / 2,
                              right: block.x + block.size / 2,
                              top: block.y - block.size / 2,
                              bottom: block.y + block.size / 2
                        })) {
        block.velx += (block.x - player.x)/80;
        block.vely += (block.y - player.y)/80;
    }
    
    for(var i = 0; i < block.motionBlur.length; i++) {
        block.motionBlur[i].age++;
    }
    
    block.motionBlur.push({
        age: 0,
        x: block.x,
        y: block.y
    });
    
    if(block.motionBlur.length > 5) {
        block.motionBlur = block.motionBlur.splice(4, 1);
    }
    
    for(var i = 0; i < enemies.array.length; i++) {
        UpdateAI(enemies.array[i]);
        
        if(enemies.array[i].dead) {
            enemies.array.splice(i, 1);
            continue;
        }
        
        if(!hasDied && RectangleCollision({
                              left: player.x - player.size / 2,
                              right: player.x + player.size / 2,
                              top: player.y - player.size / 2,
                              bottom: player.y + player.size / 2
                          }, {
                              left: enemies.array[i].x - enemies.array[i].size / 2,
                              right: enemies.array[i].x + enemies.array[i].size / 2,
                              top: enemies.array[i].y - enemies.array[i].size / 2,
                              bottom: enemies.array[i].y + enemies.array[i].size / 2
                        })) {
            hasDied = true;
            playerSound.Stop();
            playerSound.Play();
            var numParts = Math.ceil(Math.random() * 100) + 10;
            for(var j = 0; j < numParts; j++) {
                playerParticles.push({
                    x: player.x,
                    y: player.y,
                    life: 10 + Math.random() * 60,
                    velx: (Math.random() - 0.5) * 20,
                    vely: (Math.random() - 0.5) * 20,
                    size: Math.random() * 10
                });
            }
        }
    }
    
    for(var i = 0; i < particles.length; i++) {
        particles[i].life--;
        if(particles[i].life < 0) {
            particles.splice(i, 1);
            continue;
        }
        
        particles[i].x += particles[i].velx;
        particles[i].y += particles[i].vely;
    }
    
    for(var i = 0; i < playerParticles.length; i++) {
        playerParticles[i].life--;
        if(playerParticles[i].life < 0) {
            playerParticles.splice(i, 1);
            continue;
        }
        
        playerParticles[i].x += playerParticles[i].velx;
        playerParticles[i].y += playerParticles[i].vely;
    }
    
    for(var i = 0; i < enemyParticles.length; i++) {
        enemyParticles[i].life--;
        if(enemyParticles[i].life < 0) {
            enemyParticles.splice(i, 1);
            continue;
        }
        
        enemyParticles[i].x += enemyParticles[i].velx;
        enemyParticles[i].y += enemyParticles[i].vely;
    }
    
    if(hasDied && playerParticles.length == 0) {
        if(score > TOP_SCORE) {
            TOP_SCORE = score;
            SetCookie("TopScore", TOP_SCORE, 10000);
        }
        
        enemies.array = [];
        player.x = 400;
        player.y = 300;
        block.x = 400;
        block.y = 150;
        score = 0;
        hasStarted = false;
        hasDied = false;
    }
    
    shakeX *= 0.9;
    shakeY *= 0.9;
    
    if(mouseClick && RectangleCollision(mouseRect, musicRectangle)) {
        isMusicMuted = !isMusicMuted;
        if(isMusicMuted) {
            music.Stop();
        } else {
            music.Loop();
        }
    }
    
    if(mouseClick && RectangleCollision(mouseRect, soundRectangle)) {
        isSoundMuted = !isSoundMuted;
        if(isSoundMuted) {
            playerSound.SetVolume(0);
            enemySound.SetVolume(0);
            blockSound.SetVolume(0);
        } else {
            playerSound.SetVolume(0.5);
            enemySound.SetVolume(1);
            blockSound.SetVolume(1);
        }
    }
    
    mouseClick = 0;
}

function Draw(dt) {
    var sx = (Math.random() - 0.5) * shakeX;
    var sy = (Math.random() - 0.5) * shakeY;

    SpriteSetShader(SIMPLE_SHADER);
    
    if(!hasDied) {
        player.pushSprite.Draw(player.x + sx, player.y + sy, player.pushSize, player.pushSize);
        player.pullSprite.Draw(player.x + sx, player.y + sy, player.pullSize, player.pullSize);
        player.sprite.Draw(player.x + sx, player.y + sy, player.size, player.size, 0);
    }
    
    for(var i = 1; i < block.motionBlur.length; i++) {
        if(block.motionBlur[i].age >= 5) continue;
        var dx = (block.motionBlur[i].x + block.motionBlur[i-1].x) / 2 + sx;        
        var dy = (block.motionBlur[i].y + block.motionBlur[i-1].y) / 2 + sy;

        blockBlurSprites[block.motionBlur[i].age].Draw(dx, dy, block.size, block.size);
        blockBlurSprites[block.motionBlur[i].age].Draw(block.motionBlur[i].x + sx, block.motionBlur[i].y + sy, block.size, block.size);
    }
    
    block.sprite.Draw(block.x + sx, block.y + sy, block.size, block.size);
    
    for(var i = 0; i < enemies.array.length; i++) {
        enemySprite.Draw(enemies.array[i].x + sx, enemies.array[i].y + sy, enemies.array[i].size, enemies.array[i].size);
    }
    
    for(var i = 0; i < particles.length; i++) {
        particleSprite.Draw(particles[i].x, particles[i].y, particles[i].size, particles[i].size);
    }
    
    for(var i = 0; i < playerParticles.length; i++) {
        playerParticleSprite.Draw(playerParticles[i].x, playerParticles[i].y, playerParticles[i].size, playerParticles[i].size);
    }
    
    for(var i = 0; i < enemyParticles.length; i++) {
        enemyParticleSprite.Draw(enemyParticles[i].x, enemyParticles[i].y, enemyParticles[i].size, enemyParticles[i].size);
    }
    
    if(combo > 1) {
        comboText.SetText(comboStrings[combo-2]);
        comboText.Draw(280, 30);
    }
    
    if(!hasStarted) {
        text.SetText("Block pushy zoom-zoom!");
        text.Draw(260, 80);
        text.SetText("Use WASD to move, and the spacebar to push the block.");
        text.Draw(100, 470);
        text.SetText("Use the shift key to pull the block back to you.");
        text.Draw(100, 500);
        text.SetText("Watch out for the red stuff, hit them with the block.");
        text.Draw(100, 530);
        text.SetText("Push the block to start.     Mute stuff ->");
        text.Draw(250, 560);
        text.SetText(fps + " fps");
        
        if(TOP_SCORE != 0) {
            scoreText.SetText("Highscore: " + TOP_SCORE);
            scoreText.Draw(270, 30);
        }
    } else {
        currentScoreText.Draw(330, 5);
    }
    
    if(!isMusicMuted) {
        muteMusicTex.Draw(musicRectangle.left + 16, musicRectangle.top + 16, 32, 32);
    } else {
        unmuteMusicTex.Draw(musicRectangle.left + 16, musicRectangle.top + 16, 32, 32);
    }
    
    if(!isSoundMuted) {
        muteSoundTex.Draw(soundRectangle.left + 16, soundRectangle.top + 16, 32, 32);
    } else {
        unmuteSoundTex.Draw(soundRectangle.left + 16, soundRectangle.top + 16, 32, 32);
    }
    
    text.Draw(5, 5);
    frames++;
}

function RectangleCollision(r1, r2) {
  return !(r2.left > r1.right || 
           r2.right < r1.left || 
           r2.top > r1.bottom ||
           r2.bottom < r1.top);
}

function DistanceTo(x1, y1, x2, y2) {
    return Math.sqrt(Math.pow(x2 - x1, 2) + Math.pow(y2 - y1, 2));
}

function Normalize(x, y) {
    var length = Math.sqrt(x * x + y * y);
    return [x / length, y / length];
}

function Sign(x) { return x ? x < 0 ? -1 : 1 : 0; }

function CreateEnemy() {
    var side = Math.ceil(Math.random() * 4);
    var sx = 0, sy = 0;
    
    if(side == 1) { // LEFT
        sx = -100;
        sy = Math.random() * 600;
    }
    
    if(side == 2) { // RIGHT
        sx = 900;
        sy = Math.random() * 600;
    }
    
    if(side == 3) { // TOP
        sy = -100;
        sy = Math.random() * 800;
    }
    
    if(side == 4) { // BOTTOM
        sy = 700;
        sx = Math.random() * 800;
    }
    //if(!enemies) enemies = [];
    enemies.array.push({
        x: sx,
        y: sy,
        size: 32,
        speed: 2,
        dead: false,
    });
}

function BlockEnemyCollision() {
    for(var i = 0; i < enemies.array.length; i++) {
        var enemy = enemies.array[i]; 
        if(!enemy.dead) {
            if(RectangleCollision({
                                      left: enemy.x - enemy.size / 2,
                                      right: enemy.x + enemy.size / 2,
                                      top: enemy.y - enemy.size / 2,
                                      bottom: enemy.y + enemy.size / 2
                                  }, {
                                      left: block.x - block.size / 2,
                                      right: block.x + block.size / 2,
                                      top: block.y - block.size / 2,
                                      bottom: block.y + block.size / 2
                                })) {
                enemy.dead = true;
                enemySound.Stop();
                enemySound.Play();
                score += combo + 1;
                combo++;
                comboTimer = 0;
                currentScoreText.SetText("Score: " + score);
                
                var numParts = Math.ceil(Math.random() * 10) + 5;
                for(var j = 0; j < numParts; j++) {
                    particles.push({
                        x: enemy.x,
                        y: enemy.y,
                        life: 5 + Math.random() * 30,
                        velx: (Math.random() - 0.5) * 20,
                        vely: (Math.random() - 0.5) * 20,
                        size: Math.random() * 10
                    });
                }
                return;
            }
        }
    }
}

function UpdateAI(enemy) {    
    if(DistanceTo(enemy.x, enemy.y, block.x, block.y) < 40) {
        var direction = Normalize(enemy.x - block.x, enemy.y - block.y);
        enemy.x += direction[0] * enemy.speed;
        enemy.y += direction[1] * enemy.speed;
        return;
    }
    
    var direction = Normalize(enemy.x - player.x, enemy.y - player.y);
    enemy.x -= direction[0] * enemy.speed;
    enemy.y -= direction[1] * enemy.speed;
    
    enemyParticles.push({
            x: enemy.x,
            y: enemy.y,
            life: 5 + Math.random() * 20,
            velx: (Math.random() - 0.5) * 1.1,
            vely: (Math.random() - 0.5) * 1.1,
            size: Math.random() * 12
    });
}

function GetCookie(cname) {
    var name = cname + "=";
    var ca = document.cookie.split(';');
    for(var i=0; i<ca.length; i++) {
        var c = ca[i];
        while (c.charAt(0)==' ') c = c.substring(1);
        if (c.indexOf(name) == 0) return (c.substring(name.length,c.length));
    }
    return 0;
}

function SetCookie(cname, cvalue, exdays) {
    var d = new Date();
    d.setTime(d.getTime() + (exdays*24*60*60*1000));
    var expires = "expires="+d.toUTCString();
    document.cookie = cname + "=" + cvalue + "; " + expires + "; path=/";
}